#!/usr/bin/python2.7
#
#  Licensed Materials - Property of IBM
#
#  SwiftHLM Connector for IBM Spectrum Archive
#
#  (C) Copyright IBM Corp. 2016-2017 All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

"""
Authors:
Slavisa Sarafijanovic (sla@zurich.ibm.com)
Harald Seipp (seipp@de.ibm.com)
"""

"""
This file implements SwiftHLM Connector for Spectrum Archive storage backend.
On SwiftHLM side it implements SwiftHLM Generic Backend API as declared in
swifthlm_dummy_backend_connector.py of SwiftHLM. On the backend side it maps
SwiftHLM requests to the backend's migrate, recall and query status operations.
"""

from sys import stdin, stdout
from collections import defaultdict
#from swift.common import read_config_file
import ConfigParser
from swift.common.utils import readconf
from swift.common.utils import json, get_logger, split_path
import logging

from swift.obj.server import ObjectController
from swift.common.storage_policy import POLICIES
from swift.common.exceptions import ConnectionTimeout, DiskFileQuarantined, \
    DiskFileNotExist, DiskFileCollision, DiskFileNoSpace, DiskFileDeleted, \
    DiskFileDeviceUnavailable, DiskFileExpired, ChunkReadTimeout, \
    DiskFileXattrNotSupported
from swift.common.swob import HTTPAccepted, HTTPBadRequest, HTTPCreated, \
    HTTPInternalServerError, HTTPNoContent, HTTPNotFound, \
    HTTPPreconditionFailed, HTTPRequestTimeout, HTTPUnprocessableEntity, \
    HTTPClientDisconnect, HTTPMethodNotAllowed, Request, Response, \
    HTTPInsufficientStorage, HTTPForbidden, HTTPException, HeaderKeyDict, \
    HTTPConflict, HTTPServerError

import os
import errno
import uuid
import subprocess

#scor aux
from swift.proxy.controllers.base import get_container_info

#scor aux
from swift.common.utils import hash_path

regex = " || REGEXREPLACE(xattr('dmapi.IBMTPS'),['([A-Z0-9]*)@.*@([0-9]*)'],['\\1@\\2'])"
from_pool = "FROM POOL 'system'\n"
show_m_sp = "SHOW(' migrated -')\n"
show_p_sp = "SHOW(' premigrated -')\n"
show_r_sp = "SHOW(' resident -')\n"
show_m_sa = "SHOW(' migrated '" + regex + ")\n"
show_p_sa = "SHOW(' premigrated '" + regex + ")\n"
show_r_sa = "SHOW(' resident -')\n"
where_m = "WHERE (MISC_ATTRIBUTES LIKE '%V%')\n"
where_p = "WHERE (MISC_ATTRIBUTES LIKE '%M%' AND MISC_ATTRIBUTES NOT LIKE '%V%')\n"
where_r = "WHERE (NOT MISC_ATTRIBUTES LIKE '%M%')\n"

status_scan_policy_sp = \
"RULE 'MIGRATED' LIST 'status.output'\n" + from_pool + show_m_sp + where_m +\
"RULE 'PREMIGRATED' LIST 'status.output'\n" + from_pool + show_p_sp + where_p +\
"RULE 'RESIDENT' LIST 'status.output'\n" + from_pool + show_r_sp + where_r

status_scan_policy_sa = \
"RULE 'MIGRATED' LIST 'status.output'\n" + from_pool + show_m_sa + where_m +\
"RULE 'PREMIGRATED' LIST 'status.output'\n" + from_pool + show_p_sa + where_p +\
"RULE 'RESIDENT' LIST 'status.output'\n" + from_pool + show_r_sa + where_r

migrate_scan_policy_sp = \
"RULE 'MIGRATED' LIST 'mstatus.output'\n" + from_pool + show_m_sp + where_m +\
"RULE 'PREMIGRATED' LIST 'mstatus.output'\n" + from_pool + show_p_sp + where_p +\
"RULE 'RESIDENT' LIST 'mstatus.output'\n" + from_pool + show_r_sp + where_r

migrate_scan_policy_sa = \
"RULE 'MIGRATED' LIST 'mstatus.output'\n" + from_pool + show_m_sa + where_m +\
"RULE 'PREMIGRATED' LIST 'mstatus.output'\n" + from_pool + show_p_sa + where_p +\
"RULE 'RESIDENT' LIST 'mstatus.output'\n" + from_pool + show_r_sa + where_r

recall_scan_policy_sp = \
"RULE 'MIGRATED' LIST 'rstatus.output'\n" + from_pool + show_m_sp + where_m +\
"RULE 'PREMIGRATED' LIST 'rstatus.output'\n" + from_pool + show_p_sp + where_p +\
"RULE 'RESIDENT' LIST 'rstatus.output'\n" + from_pool + show_r_sp + where_r

recall_scan_policy_sa = \
"RULE 'MIGRATED' LIST 'rstatus.output'\n" + from_pool + show_m_sa + where_m +\
"RULE 'PREMIGRATED' LIST 'rstatus.output'\n" + from_pool + show_p_sa + where_p +\
"RULE 'RESIDENT' LIST 'rstatus.output'\n" + from_pool + show_r_sa + where_r

#BACKEND = 'ibmsa'
BACKEND = 'ibmsp'
#BACKEND = 'ibmsp-dsmls'

# SwiftHLM Backend Connector
class SwiftHlmBackendConnector(object):

    def __init__(self):
        self.__request_in = {}
        self.__request_out = {}
        self.__request_out_request = ''
        self.__request_out_filelist = ''
        self.__response_in = {}
        self.__response_out = {}

        # Config
        configFile = r'/etc/swift/object-server.conf'
        self.conf = readconf(configFile) 

        # Logging
        self.hlm_stor_node_config = self.conf.get('hlm', None)
        if self.hlm_stor_node_config:
            hlm_stor_node_log_level = self.hlm_stor_node_config.get( \
                    'set log_level', None)
        if hlm_stor_node_log_level:
            self.conf['log_level'] = hlm_stor_node_log_level
        self.logger = get_logger(self.conf, name='hlm-connector',
                log_route='swifthlm', fmt="%(server)s: %(msecs)03d "
                "[%(filename)s:%(funcName)20s():%(lineno)s] %(message)s")

        self.gbi_provide_dirpaths_instead_of_filepaths = False
        conf_gbi_provide_dirpaths_instead_of_filepaths = \
            self.hlm_stor_node_config.get( \
            'gbi_provide_dirpaths_instead_of_filepaths', 'False')
        if conf_gbi_provide_dirpaths_instead_of_filepaths == 'True':    
            self.gbi_provide_dirpaths_instead_of_filepaths = True

        self.logger.info('info: Initialized Connector')
        self.logger.debug('dbg: Initialized Connector')
        self.logger.info('conf: %s', json.dumps(self.conf))

    # Next method is to be invoked by SwiftHLM Handler using SwiftHLM Generic
    # Backend Interface (GBI). It adapts SwiftHLM request for IBM Spectrum
    # Protect storage backend, invokes the backend operations, reformats the
    # backend response to GBI format, and returns the response to SwitHLM
    # Handler
    def submit_request_get_response(self, request):
        self.__receive_request(request)
        self.__reformat_swifthlm_request_to_specific_backend_api()
        self.__submit_request_to_backend_get_response()
        self.__reformat_backend_response_to_generic_backend_api()
        return self.__response_out

    # This method receives the request from SwiftHLM Handler
    def __receive_request(self, request):
        
        self.logger.debug('Receiving request from Handler')
        self.__request_in = request

        return
       
    # This method reformats request to backend API
    def __reformat_swifthlm_request_to_specific_backend_api(self):

        self.logger.debug('Reformatting request to the specific Backend API')
        self.logger.debug('request_in(first 1024 bytes): %s',
                str(self.__request_in)[0:1023])
    
        # Backend specific part
        self.__request_out_request = self.__request_in['request']
        if str.lower(self.__request_in['request']) == 'status':
            # status: reuse input request as is
            self.__request_out = self.__request_in
        else:
            # migration or recall: prepare list for bulk migration/recall
            # make sure temporary directory is configured and created
            cfg = self.read_connector_and_backend_config(
                    '/etc/swift/object-server.conf')
            connector_tmp_dir = cfg['connector_tmp_dir']
            if connector_tmp_dir:
                self.mkdir_minus_p(connector_tmp_dir)    
            else:
                self.logger.error('swifthlm temporary directory not configured')
                raise
            # temporary list file name
            tmp_filename = str(uuid.uuid1())
            self.__request_out_list = connector_tmp_dir + '/' + tmp_filename
            f = open(self.__request_out_list, 'w')
            for obj_and_file in self.__request_in['objects']: 
                f.write(' -- ' + str(obj_and_file['file']) + '\n')
            f.close()
            fr = open(self.__request_out_list, 'r')
            file_list_content = fr.read()
            self.logger.debug('file_list: %s', file_list_content)
            fr.close()

        return

    # This method submits request to Backend and gets Response from Backend 
    def __submit_request_to_backend_get_response(self):
        
        self.logger.debug('Submitting request to backend')
        if self.__request_out_request == 'status':
            # query status
            #self.query_status_receive_response()
            self.query_status_receive_response2()
        elif self.__request_out_request == 'migrate':
            #self.__response_in = 0
            #self.migrate_receive_response()
            self.migrate_receive_response2()
        elif self.__request_out_request == 'recall':
            #self.recall_receive_response()
            self.recall_receive_response2()
        else: #wrong request, TODO: move this check, do early as possible
            raise    
        return

    def __reformat_backend_response_to_generic_backend_api(self):

        self.logger.debug('Reformatting response to Generic Backend API')
        self.logger.debug('response_in(first 1024 bytes): %s',
                str(self.__response_in)[0:1023])
    
        # Backend specific part, for the assumed dummy backend it just copies the
        # incoming response from the backend
        self.__response_out = self.__response_in

        return

    def mkdir_minus_p(self, dir_path):
        try:
            os.makedirs(dir_path)
        except OSError as err: # TODO: check python 3.x
            if err.errno == errno.EEXIST and os.path.isdir(dir_path):
                pass
            else:
                raise
        return

    def migrate_receive_response2(self):
        # TODO: consider option to pass directories, not files, in order to
        # save work in handler and make overall operation faster
        # make sure temporary directory is configured and created
        self.logger.debug('migrate_receive_response2()')
        request = self.__request_out_request
        cfg = self.read_connector_and_backend_config(
                '/etc/swift/object-server.conf')
        connector_tmp_dir = cfg['connector_tmp_dir']
        if connector_tmp_dir:
            self.mkdir_minus_p(connector_tmp_dir)
        else:
            self.logger.error('swifthlm temporary directory not configured')
            raise

        self.logger.debug('prepare migration policy file')
        # prepare migration scan policy file
        #scan_policy_file = connector_tmp_dir + "/migpolicy"
        #scan_policy_file = connector_tmp_dir + "/regexpolicy"
        scan_policy_file = connector_tmp_dir + "/mregexpolicy"
        if not os.path.isfile(scan_policy_file):
            f = open(scan_policy_file, 'w')
            if BACKEND == "ibmsp":
               f.write(migrate_scan_policy_sp)
            else:
               f.write(migrate_scan_policy_sa)
            f.close()

        # prepare temporary lists unique file name prefix
        lists_prefix = str(uuid.uuid1())
        input_list = connector_tmp_dir + '/' + lists_prefix \
                                            + '.list.mstatus.input'
        self.logger.debug('input_list: %s', input_list)
        f = open(input_list, 'w')
        for obj_and_file in self.__request_in['objects']:
            f.write(str(obj_and_file['file']) + '\n')
        f.close()

        # mmapplypolicy output is by default owned by root, 0600 file mode
        # so we create it as swift user to be able to process it later
        output_list = connector_tmp_dir + '/' + lists_prefix \
                                            + '.list.mstatus.output'
        self.logger.debug('output_list: %s', output_list)
        open(output_list, 'w').close()
        output_list_prefix = connector_tmp_dir + '/' + lists_prefix

        # Prepare scan command
        cmd = ["sudo",
                "/usr/lpp/mmfs/bin/mmapplypolicy",
                cfg['gpfs_filesystem_or_fileset'],
                "-P" + scan_policy_file,
                "-i" + input_list,
                "-Idefer",
                "-f" + output_list_prefix]

        # Invoke scan command
        self.logger.info('iInvoke mmapplypolicy scan.')
        self.logger.debug('Invoke mmapplypolicy scan.')
        p = subprocess.Popen(cmd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
        out, err = p.communicate()
        self.logger.info('iRunning mmapplypolicy scan completed.')
        self.logger.debug('Running mmapplypolicy scan completed.')
        # Check result
        if p.returncode:
            self.logger.error('Migration scan errors: %s', err)
            return
        
        fr = open(output_list, 'r')
        file_list_content = fr.read()
        self.logger.debug('output_list(first 1024 bytes): %s', 
                        str(file_list_content)[0:1023])
        fr.close()

        # Reformat the list for ltfsee/dsmmigrate 
        listfile = output_list + '.filenames'
        miglist_empty = True
        fr = open(output_list, 'r')
        f = open(listfile, 'w')
        for line in fr.readlines():
            file_status = line.split()[3]     
            file_path = line.split()[-1]
            if file_status == 'resident' or file_status == 'premigrated':
                if BACKEND == 'ibmsa':
                    f.write(' -- ' + file_path + '\n')
                    miglist_empty = False
                else:
                    f.write(file_path + '\n')
                    miglist_empty = False
            elif file_status == 'r' or file_status == 'p':
                f.write(file_path + '\n')
                miglist_empty = False
        fr.close()
        f.close()

        # If all files already migrated return success
        if miglist_empty:
            self.__response_in = 0
            return

        # Migrate object files - unfortunately migrate commands must be run as
        # root
        if BACKEND == 'ibmsa':
            cmd = ["sudo", "/opt/ibm/ltfsee/bin/ltfsee", "migrate",
                    "-s", listfile, '-p']
            for pool in cfg['tape_storage_pool'].split():
                cmd.append(pool)
        elif 'ibmsp' in BACKEND: 
            cmd = ["sudo", "/usr/bin/dsmmigrate", "-FIlelist=" + listfile]
        p = subprocess.Popen(cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE)
        out, error = p.communicate()
        rc = p.returncode
        self.logger.debug('migrate.out(first 1024 bytes): %s',
                str(out)[0:1023]) 
        if rc == 6:
            rc = 0
        self.__response_in = rc
        return

    def migrate_receive_response(self):
        listfile = self.__request_out_list
        request = self.__request_out_request
        cfg = self.read_connector_and_backend_config(
                '/etc/swift/object-server.conf')
        # TODO: read migration program path (sa vs sp) from config file
        # for now fixed value "opt/ibm/ltfsee/bin/ltfsee" is used
        # migrate object files - unfortunately ltfsee migrate must be run as
        # root
        if BACKEND == 'ibmsa':
            cmd = ["sudo", "/opt/ibm/ltfsee/bin/ltfsee", "migrate",
                    "-s", listfile, '-p']
            for pool in cfg['tape_storage_pool'].split():
                cmd.append(pool)
        elif 'ibmsp' in BACKEND: 
            cmd = ["sudo", "/usr/bin/dsmmigrate", "-FIlelist=" + listfile]
        p = subprocess.Popen(cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE)
        out, error = p.communicate()
        rc = p.returncode
        self.logger.debug('migrate.out(first 1024 bytes): %s',
                str(out)[0:1023]) 
        if rc == 6:
            rc = 0
        self.__response_in = rc
        return

    def recall_receive_response2(self):
        # TODO: consider option to pass directories, not files, in order to
        # save work in handler and make overall operation faster
        # make sure temporary directory is configured and created
        self.logger.debug('recall_receive_response2()')
        request = self.__request_out_request
        cfg = self.read_connector_and_backend_config(
                '/etc/swift/object-server.conf')
        connector_tmp_dir = cfg['connector_tmp_dir']
        if connector_tmp_dir:
            self.mkdir_minus_p(connector_tmp_dir)
        else:
            self.logger.error('swifthlm temporary directory not configured')
            raise

        self.logger.debug('prepare recall policy file')
        # prepare migration scan policy file
        #scan_policy_file = connector_tmp_dir + "/regexpolicy"
        scan_policy_file = connector_tmp_dir + "/rregexpolicy"
        if not os.path.isfile(scan_policy_file):
            f = open(scan_policy_file, 'w')
            if BACKEND == "ibmsp":
               f.write(recall_scan_policy_sp)
            else:
               f.write(recall_scan_policy_sa)
            f.close()

        # prepare temporary lists unique file name prefix
        lists_prefix = str(uuid.uuid1())
        input_list = connector_tmp_dir + '/' + lists_prefix \
                                            + '.list.rstatus.input'
        self.logger.debug('input_list: %s', input_list)
        f = open(input_list, 'w')
        for obj_and_file in self.__request_in['objects']:
            f.write(str(obj_and_file['file']) + '\n')
        f.close()

        # mmapplypolicy output is by default owned by root, 0600 file mode
        # so we create it as swift user to be able to process it later
        output_list = connector_tmp_dir + '/' + lists_prefix \
                                            + '.list.rstatus.output'
        self.logger.debug('output_list: %s', output_list)
        open(output_list, 'w').close()
        output_list_prefix = connector_tmp_dir + '/' + lists_prefix

        # Prepare scan command
        cmd = ["sudo",
                "/usr/lpp/mmfs/bin/mmapplypolicy",
                cfg['gpfs_filesystem_or_fileset'],
                "-P" + scan_policy_file,
                "-i" + input_list,
                "-Idefer",
                "-f" + output_list_prefix]

        # Invoke scan command
        self.logger.info('iInvoke mmapplypolicy scan.')
        self.logger.debug('Invoke mmapplypolicy scan.')
        p = subprocess.Popen(cmd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
        out, err = p.communicate()
        self.logger.info('iRunning mmapplypolicy scan completed.')
        self.logger.debug('Running mmapplypolicy scan completed.')
        # Check result
        if p.returncode:
            self.logger.error('Recall scan errors: %s', err)
            return
        
        fr = open(output_list, 'r')
        file_list_content = fr.read()
        self.logger.debug('output_list(first 1024 bytes): %s', 
                        str(file_list_content)[0:1023])
        fr.close()

        # Reformat the list for ltfsee/dsmrecall
        listfile = output_list + '.filenames'
        miglist_empty = True
        fr = open(output_list, 'r')
        f = open(listfile, 'w')
        for line in fr.readlines():
            file_status = line.split()[3]     
            file_path = line.split()[-1]
            if file_status == 'migrated':
               if BACKEND == 'ibmsa':
                   f.write(' -- ' + file_path + '\n')
                   miglist_empty = False
               else:
                   f.write(file_path + '\n')
                   miglist_empty = False
            elif file_status == 'm':
                f.write(file_path + '\n')
                miglist_empty = False


        fr.close()
        f.close()

        # If all files already recalled return success
        if miglist_empty:
            self.__response_in = 0
            return

        # Recall object files - unfortunately recall commands must be run as
        # root
        if BACKEND == 'ibmsa':
            cmd = ["sudo", "/opt/ibm/ltfsee/bin/ltfsee", "recall", 
                    "-l", cfg['library'], "-s", listfile]
        elif 'ibmsp' in BACKEND:
            cmd = ["sudo", "/usr/bin/dsmrecall", "-FIlelist=" + listfile, cfg['gpfs_filesystem_or_fileset']]
        p = subprocess.Popen(cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE)
        out, error = p.communicate()
        rc = p.returncode
        self.logger.debug('recall.out(first 1024 bytes): %s', str(out)[0:1023]) 
        self.__response_in = rc
        return

    def recall_receive_response(self):
        listfile = self.__request_out_list
        request = self.__request_out_request
        cfg = self.read_connector_and_backend_config(
                '/etc/swift/object-server.conf')
        # TODO: read migration program path (sa vs sp) from config file
        # for now fixed value "opt/ibm/ltfsee/bin/ltfsee" is used
        # migrate object files - unfortunately ltfsee migrate must be run as
        # root
        if BACKEND == 'ibmsa':
            cmd = ["sudo", "/opt/ibm/ltfsee/bin/ltfsee", "recall", 
                    "-l", cfg['library'], "-s", listfile]
        elif 'ibmsp' in BACKEND:
            cmd = ["sudo", "/usr/bin/dsmrecall", "-FIlelist=" + listfile, cfg['gpfs_filesystem_or_fileset']]
        p = subprocess.Popen(cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE)
        out, error = p.communicate()
        rc = p.returncode
        self.logger.debug('recall.out(first 1024 bytes): %s', str(out)[0:1023]) 
        self.__response_in = rc
        return

    def query_status_receive_response2(self):
        # TODO: consider option to pass directories, not files, in order to
        # save work in handler and make overall operation faster
        # make sure temporary directory is configured and created
        self.logger.debug('query_status_receive_response')
        cfg = self.read_connector_and_backend_config(
                '/etc/swift/object-server.conf')
        connector_tmp_dir = cfg['connector_tmp_dir']
        if connector_tmp_dir:
            self.mkdir_minus_p(connector_tmp_dir)
        else:
            self.logger.error('swifthlm temporary directory not configured')
            raise

        self.logger.debug('prepare scan policy file')
        # prepare status scan policy file
        scan_policy_file = connector_tmp_dir + "/regexpolicy"
        if not os.path.isfile(scan_policy_file):
            f = open(scan_policy_file, 'w')
            if BACKEND == "ibmsp":
               f.write(status_scan_policy_sp)
            else:
               f.write(status_scan_policy_sa)
            f.close()

        # prepare temporary lists unique file name prefix
        lists_prefix = str(uuid.uuid1())
        input_list = connector_tmp_dir + '/' + lists_prefix \
                                            + '.list.status.input'
        self.logger.debug('input_list: %s', input_list)
        f = open(input_list, 'w')
        for obj_and_file in self.__request_in['objects']:
            f.write(str(obj_and_file['file']) + '\n')
        f.close()

        # mmapplypolicy output is by default owned by root, 0600 file mode
        # so we create it as swift user to be able to process it later
        output_list = connector_tmp_dir + '/' + lists_prefix \
                                            + '.list.status.output'
        self.logger.debug('output_list: %s', output_list)
        open(output_list, 'w').close()
        output_list_prefix = connector_tmp_dir + '/' + lists_prefix

        if BACKEND == 'ibmsa' or BACKEND == 'ibmsp':
            # Prepare status scan command
            cmd = ["sudo",
                    "/usr/lpp/mmfs/bin/mmapplypolicy",
                    cfg['gpfs_filesystem_or_fileset'],
                    "-P" + scan_policy_file,
                    "-i" + input_list,
                    "-Idefer",
                    "-f" + output_list_prefix]
        elif BACKEND == 'ibmsp-dsmls': #TODO rm if scan sol. works for ibmsp
            # Prepare dsmls command to get files status
            cmd = ["sudo" + \
                    " /usr/bin/dsmls" + \
                    " -Noheader" + \
                    " -FIlelist=" + input_list + \
                    " | awk 'NR > 6 { print }'" + \
                    " >" + output_list]
#            cmd = ["sudo",
#                    "/usr/bin/dsmls",
#                    "-Noheader",
#                    "-FIlelist=" + input_list]
        # Invoke the command
        self.logger.info('iInvoke mmapplypolicy.')
        self.logger.debug('Invoke mmapplypolicy.')
        if BACKEND == 'ibmsa' or BACKEND == 'ibmsp':
            p = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        elif BACKEND == 'ibmsp-dsmls': #TODO rm if scan sol. works for ibmsp
            p = subprocess.Popen(cmd,
                                shell=True,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = p.communicate()
        self.logger.info('iRunning mmapplypolicy completed.')
        self.logger.debug('Running mmapplypolicy completed.')
        # Check result
        if p.returncode:
            self.logger.error('Status query errors: %s', err)
            return

        fr = open(output_list, 'r')
        file_list_content = fr.read()
        self.logger.debug('output_list(first 1024 bytes): %s', 
                        str(file_list_content)[0:1023])
        fr.close()

        # get file-status pairs
        self.logger.debug('Get file-status pairs')
        filenames_statuses = {} # for dsmls, rm if scan solution works
        dirpaths_filepaths = {}
        dirpaths_statuses = {}
        fr = open(output_list, 'r')
        for line in fr.readlines():
            file_path = line.split()[-1]
            file_status = line.split()[3]     
            if BACKEND == 'ibmsa' or BACKEND == 'ibmsp':
                if self.gbi_provide_dirpaths_instead_of_filepaths:
                    dir_path = os.path.dirname(file_path)
                else: # TODO dir_path -> dirpath_or_filepath
                    dir_path = file_path
                dirpaths_filepaths[dir_path] = file_path 
                dirpaths_statuses[dir_path] = file_status
            elif BACKEND == 'ibmsp-dsmls': # TODO rm if scan solution works 
                if file_status == 'r':
                    file_status = 'resident'
                elif file_status == 'p':
                    file_status = 'premigrated'
                elif file_status == 'm':
                    file_status = 'migrated'
                filenames_statuses[file_name] = file_status    

        # create object to dir_or_file to status mapping 
        self.logger.debug('Create object to file to status mapping')
        objects = []
        for obj_and_file in self.__request_out['objects']:
            obj_file_status = {}
            obj_file_status['object'] = obj_and_file['object']
            #obj_file_status['file'] = obj_and_file['file']
            if BACKEND == 'ibmsa' or BACKEND == 'ibmsp':
                dirpathkey = obj_and_file['file'] # TODO 'file'->'dirorfilepath
                obj_file_status['file'] = dirpaths_filepaths[dirpathkey]
                obj_file_status['status'] = dirpaths_statuses[dirpathkey]
            elif BACKEND == 'ibmsp-dsmls':
                obj_file_status['file'] = obj_and_file['file']
                filenamekey = obj_and_file['file'].split('/')[-1]
                obj_file_status['status'] = filenames_statuses[filenamekey]
            objects.append(obj_file_status)
        self.__response_in['objects'] = objects

        # clean up TODO: uncomment next 2 lines once major defects are fixed
        #os.remove(input_list)
        #os.remove(output_list)
        self.logger.debug('Return from query_status_receive_response()')
        return

    def query_status_receive_response(self):
        # prepare
        objects = []
        rc = 0
        files_str = ''
        files = []
        for obj_and_file in self.__request_out['objects']:
            files_str += obj_and_file['file'] + ' '
            files.append(obj_and_file['file'])
        # query backend    
        p = subprocess.Popen(["/usr/bin/xargs",
            "/opt/ibm/ltfsee/bin/ltfsee", "info", "files", "-q", "-f"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE)
        out, error = p.communicate(" ".join(map(str,files)))
        rc = p.returncode
        self.logger.debug('out(first 1024 bytes): %s', str(out)[0:1023])
        # process response
        if error:
            self.logger.error('Status query errors: %s', error)
        if rc != 0:
            self.set_statuses_to_unknown()
            return
        lines = out.split('\n')
        n_lines = len(lines)
        self.logger.debug('n_lines: %s', n_lines)
        #if not out or len(lines) % 2:
        if not out:
            self.logger.error('Status query unexpected output: %s', out)
            self.set_statuses_to_unknown()        
            return
        # get file-status pairs
        names_statuses = {}
        i=0
        while i < n_lines-2:
            # file-line
            file_line = lines[i]
            if file_line.startswith("Name: "):
                file_name = file_line[6:]
            else:
                self.logger.error('Status query unexpected name-line: %s',
                        file_line)
                self.set_statuses_to_unknown()
                return
            #info-line
            info_line = lines[i+1]
            infos = info_line.split()
            self.logger.debug('infos: %s', str(infos))
            infos_len = len(infos)
            if infos_len < 4 or infos[infos_len-2] != 'Status:':
                self.logger.error('Status query unexpected info-line: %s',
                        info_line)
                self.set_statuses_to_unknown()
                return
            else:
                file_status = infos[infos_len-1] 
            names_statuses[file_name] = file_status
            i += 2
        # create object to file to status mapping
        objects = []
        for obj_and_file in self.__request_out['objects']:
            obj_file_status = {}
            obj_file_status['object'] = obj_and_file['object']
            obj_file_status['file'] = obj_and_file['file']
            obj_file_status['status'] = names_statuses[obj_and_file['file']]
            objects.append(obj_file_status)
        self.__response_in['objects'] = objects
        return

    def set_statuses_to_unknown(self):
        objects = []
        for obj_and_file in self.__request_out['objects']:
            obj_file_status = obj_and_file
            obj_file_status['status'] = 'unknown'
            objects.append(obj_file_status)
        self.__response_in['objects'] = objects
        return

    def read_connector_and_backend_config(self, configfile_path):
        configParser = ConfigParser.RawConfigParser()
        configParser.read(configfile_path)
        config = {}
        config['gpfs_filesystem_or_fileset'] = configParser.get('ibmsasp',
                'gpfs_filesystem_or_fileset')
        config['library'] = configParser.get('ibmsasp', 'library')
        config['tape_storage_pool'] = configParser.get('ibmsasp',
                'tape_storage_pool')
        config['connector_tmp_dir'] = configParser.get('ibmsasp',
                'connector_tmp_dir') 
        return config


if __name__ == '__main__':
    # SwiftHlmConnector class is not assumed to be used standalone, instead it
    # is imported for a configured backend by SwiftHLM Handler and invoked from
    # the Handler. Alternatively it could be modified to be invoked as a new
    # process and/or remoteley similar to SwiftHLM Dispatcher invoking SwiftHLM
    # Handler
    raise 

